# fit seq
# fitting tscan data
# Using sequential decay model convolved with
# normalized gaussian distribution
# normalized cauchy distribution
# normalized pseudo voigt profile
# (Mixing parameter eta is fixed according to
#  Journal of Applied Crystallography. 33 (6): 1311–1316.)

import argparse
import numpy as np
from ..mathfun.rate_eq import compute_signal_irf, fact_anal_model
from ..mathfun import solve_seq_model, rate_eq_conv, fact_anal_rate_eq_conv
from .misc import set_bound_tau, read_data, contribution_table, plot_result
from lmfit import Parameters, fit_report, minimize

description = '''
fit seq: fitting tscan data using the solution of sequtial decay equation covolved with gaussian/cauchy(lorenzian)/pseudo voigt irf function.
It uses lmfit python module to fitting experimental time trace data to sequential decay module.
To find contribution of each excited state species, it solves linear least square problem via scipy lstsq module.


It supports 4 types of sequential decay
Type 0: both raising and decay
    GS -> 1 -> 2 -> ... -> n -> GS
Type 1: no raising
    1 -> 2 -> ... -> n -> GS
Type 2: no decay
    GS -> 1 -> 2 -> ... -> n
Type 3: Neither raising nor decay
    1 -> 2 -> ... -> n 
'''

epilog = '''
*Note
1. The number of time zero parameter should be same as the number of scan to fit.
2. Type 0 sequential decay needs n+1 lifetime constants for n excited state species
3. Type 1, 2 sequential decay needs n lifetime constants for n excited state species
4. Type 3 sequential decay needs n-1 lifetime constants for n excited state species
5. if you set shape of irf to pseudo voigt (pv), then you should provide two full width at half maximum value for gaussian and cauchy parts, respectively.
'''

seq_decay_help = '''
type of sequential decay
0. GS -> 1 -> 2 -> ... -> n -> GS (both raising and decay)
1. 1 -> 2 -> ... -> n -> GS (No raising)
2. GS -> 1 -> 2 -> ... -> n (No decay)
3. 1 -> 2 -> ... -> n (Neither raising nor decay)
Default option is type 0 both raising and decay

*Note
1. type 0 needs n+1 lifetime value
2. type 1 and 2 need n lifetime value
3. type 3 needs n-1 lifetime value
'''

irf_help = '''
shape of instrument response functon
g: gaussian distribution
c: cauchy distribution
pv: pseudo voigt profile, linear combination of gaussian distribution and cauchy distribution 
    pv = eta*c+(1-eta)*g 
    the mixing parameter is fixed according to Journal of Applied Crystallography. 33 (6): 1311–1316. 
'''

fwhm_G_help = '''
full width at half maximum for gaussian shape
It would not be used when you set cauchy irf function
'''

fwhm_L_help = '''
full width at half maximum for cauchy shape
It would not be used when you did not set irf or use gaussian irf function
'''



def fit_seq():

    def residual(params, t, num_tau, exclude, irf, data=None, eps=None):
        if irf in ['g', 'c']:
            fwhm = params['fwhm']
        else:
            fwhm = np.array([params['fwhm_G'], params['fwhm_L']])
        tau = np.zeros(num_tau)
        for i in range(num_tau):
            tau[i] = params[f'tau_{i+1}']
        eigval, V, c = solve_seq_model(tau)

        chi = np.zeros((data.shape[0], data.shape[1]))
        for i in range(data.shape[1]):
            t0 = params[f't_0_{i+1}']
            model = compute_signal_irf(t-t0, eigval, V, c, fwhm, irf)
            abs = fact_anal_model(model, exclude, data[:, i], eps[:, i])
            chi[:, i] = data[:, i] - (abs @ model)
        chi = chi.flatten()/eps.flatten()

        return chi

    tmp = argparse.RawTextHelpFormatter
    parser = argparse.ArgumentParser(formatter_class=tmp,
                                     description=description,
                                     epilog=epilog)
    parser.add_argument('-sdt', '--seq_decay_type', type=int, default=0,
                        help=seq_decay_help)
    parser.add_argument('--irf', default='g', choices=['g', 'c', 'pv'],
                        help=irf_help)
    parser.add_argument('--fwhm_G', type=float,
                        help=fwhm_G_help)
    parser.add_argument('--fwhm_L', type=float,
                        help=fwhm_L_help)
    parser.add_argument('prefix',
                        help='prefix for tscan files ' +
                        'It will read prefix_i.txt')
    parser.add_argument('-t0', '--time_zeros', type=float, nargs='+',
                        help='time zeros for each tscan')
    parser.add_argument('-t0f', '--time_zeros_file',
                        help='filename for time zeros of each tscan')
    parser.add_argument('--tau', type=float, nargs='*',
                        help='lifetime of each decay')
    parser.add_argument('--fix_irf', action='store_true',
    help='fix irf parameter (fwhm_G, fwhm_L) during fitting process')
    parser.add_argument('--slow', action='store_true',
    help='use slower but robust global optimization algorithm')
    parser.add_argument('-o', '--out', default=None,
                        help='prefix for output files')
    args = parser.parse_args()

    prefix = args.prefix
    if args.out is None:
        args.out = prefix
    out_prefix = args.out

    seq_decay_type = args.seq_decay_type

    irf = args.irf
    if irf == 'g':
        if args.fwhm_G is None:
            print('You are using gaussian irf, so you should set fwhm_G!\n')
            return
        else:
            fwhm = args.fwhm_G
    elif irf == 'c':
        if args.fwhm_L is None:
            print('You are using cauchy/lorenzian irf,' +
                  'so you should set fwhm_L!\n')
            return
        else:
            fwhm = args.fwhm_L
    else:
        if (args.fwhm_G is None) or (args.fwhm_L is None):
            print('You are using pseudo voigt irf,' +
                  'so you should set both fwhm_G and fwhm_L!\n')
            return
        else:
            fwhm = 0.5346*args.fwhm_L + \
                np.sqrt(0.2166*args.fwhm_L**2+args.fwhm_G**2)

    if args.tau is None:
        print('Please set lifetime constants for each decay')
        return
    else:
        tau = np.array(args.tau)
        num_tau = tau.size
        if seq_decay_type == 0:
            num_ex = num_tau-1
            exclude = 'first_and_last'
        elif seq_decay_type == 1:
            num_ex = num_tau
            exclude = 'last'
        elif seq_decay_type == 2:
            num_ex = num_tau
            exclude = 'first'
        else:
            num_ex = num_tau+1
            exclude = None
    
    if (args.time_zeros is None) and (args.time_zeros_file is None):
        print('You should set either time_zeros or time_zeros_file!\n')
        return
    elif args.time_zeros is None:
        time_zeros = np.genfromtxt(args.time_zeros_file)
        num_scan = time_zeros.size
    else:
        time_zeros = np.array(args.time_zeros)
        num_scan = time_zeros.size

    t = np.genfromtxt(f'{prefix}_1.txt')[:, 0]
    num_data_pts = t.size
    data, eps = read_data(prefix, num_scan, num_data_pts, 10)

    print(f'fitting with {num_scan} data set!\n')
    fit_params = Parameters()
    if irf in ['g', 'c']:
        fit_params.add('fwhm', value=fwhm,
                       min=0.5*fwhm, max=2*fwhm, vary=(not args.fix_irf))
    elif irf == 'pv':
        fit_params.add('fwhm_G', value=args.fwhm_G,
                       min=0.5*args.fwhm_G, max=2*args.fwhm_G, vary=(not args.fix_irf))
        fit_params.add('fwhm_L', value=args.fwhm_L,
                       min=0.5*args.fwhm_L, max=2*args.fwhm_L, vary=(not args.fix_irf))
    for i in range(num_scan):
        fit_params.add(f't_0_{i+1}', value=time_zeros[i],
                       min=time_zeros[i]-2*fwhm,
                       max=time_zeros[i]+2*fwhm)

    for i in range(num_tau):
        bd = set_bound_tau(tau[i])
        fit_params.add(f'tau_{i+1}', value=tau[i], min=bd[0], max=bd[1])

    # Second initial guess using global optimization algorithm
    if args.slow: 
        out = minimize(residual, fit_params, method='ampgo', calc_covar=False,
        args=(t, num_tau, exclude, irf),
        kws={'data': data, 'eps': eps})
    else:
        out = minimize(residual, fit_params, method='nelder', calc_covar=False,
        args=(t, num_tau, exclude, irf),
        kws={'data': data, 'eps': eps})

    # Then do Levenberg-Marquardt
    out = minimize(residual, out.params,
                   args=(t, num_tau, exclude),
                   kws={'data': data, 'eps': eps, 'irf': irf})

    chi2_ind = residual(out.params, t, num_tau, exclude,
                        irf, data=data, eps=eps)
    chi2_ind = chi2_ind.reshape(data.shape)
    chi2_ind = np.sum(chi2_ind**2, axis=0)/(data.shape[0]-len(out.params))

    fit = np.zeros((data.shape[0], data.shape[1]+1))
    res = np.zeros((data.shape[0], data.shape[1]+1))
    fit[:, 0] = t
    res[:, 0] = t

    if irf in ['g', 'c']:
        fwhm_opt = out.params['fwhm']
    else:
        tmp_G = out.params['fwhm_G']
        tmp_L = out.params['fwhm_L']
        fwhm_opt = np.array([tmp_G, tmp_L])

    tau_opt = np.zeros(num_tau)
    for j in range(num_tau):
        tau_opt[j] = out.params[f'tau_{j+1}']
    
    eigval_opt, V_opt, c_opt = solve_seq_model(tau_opt)

    abs = np.zeros((num_ex, num_scan))
    for i in range(num_scan):
        abs_tmp = fact_anal_rate_eq_conv(t-out.params[f't_0_{i+1}'],
        fwhm_opt, eigval_opt, V_opt, c_opt, exclude, 
        data=data[:, i], eps=eps[:, i], irf=irf)
        fit[:, i+1] = rate_eq_conv(t-out.params[f't_0_{i+1}'],
        fwhm_opt, abs_tmp, eigval_opt, V_opt, c_opt, irf=irf)
        if seq_decay_type == 0:
            abs[:, i] = abs_tmp[1:-1]
        elif seq_decay_type == 1:
            abs[:, i] = abs_tmp[:-1]
        elif seq_decay_type == 2:
            abs[:, i] = abs_tmp[1:]
        else:
            abs[:, i] = abs_tmp
    
    res[:, 1:] = data - fit[:, 1:]
    
    contrib_table = contribution_table('tscan', 'Excited State Contribution', num_scan,
    num_ex, abs)
    fit_content = fit_report(out) + '\n' + contrib_table

    print(fit_content)

    f = open(out_prefix+'_fit_report.txt', 'w')
    f.write(fit_content)
    f.close()

    np.savetxt(out_prefix+'_fit.txt', fit)
    np.savetxt(out_prefix+'_abs.txt', abs)

    # save residual of individual fitting 

    for i in range(data.shape[1]):
        res_ind = np.vstack((res[:, 0], res[:, i+1], eps[:, i]))
        np.savetxt(out_prefix+f'_res_{i+1}.txt', res_ind.T)

    plot_result('tscan', num_scan, chi2_ind, data, eps, fit, res)

    return



'''
exp_conv_irf:
submodule for the mathematical functions for
irf (instrumental response function)

:copyright: 2022 by pistack (Junho Lee).
:license: LGPL3.
'''

from typing import Union
import numpy as np

def gau_irf(t: Union[float, np.ndarray], fwhm: float) -> Union[float, np.ndarray]:

    '''
    Compute gaussian shape irf function

    Args:
      t: time
      fwhm: full width at half maximum of gaussian distribution

    Returns:
     normalized gaussian function.
    '''

    sigma = fwhm/(2*np.sqrt(2*np.log(2)))
    return np.exp(-t**2/(2*sigma**2))/(sigma*np.sqrt(2*np.pi))

def cauchy_irf(t: Union[float, np.ndarray], fwhm: float) -> Union[float, np.ndarray]:

    '''
    Compute lorenzian shape irf function

    Args:
      t: time
      fwhm: full width at half maximum of cauchy distribution

    Returns:
     normalized lorenzian function.
    '''

    gamma = fwhm/2
    return gamma/np.pi/(t**2+gamma**2)

def pvoigt_irf(t: Union[float, np.ndarray], fwhm_G: float, fwhm_L: float, eta: float) -> Union[float, np.ndarray]:
    '''
    Compute pseudo voight shape irf function
    (i.e. linear combination of gaussian and lorenzian function)

    Args:
      t: time
      fwhm_G: full width at half maximum of gaussian part
      fwhm_L: full width at half maximum of lorenzian part
      eta: mixing parameter
    Returns:
     linear combination of gaussian and lorenzian function with mixing parameter eta.
    '''

    u = gau_irf(t, fwhm_G)
    v = cauchy_irf(t, fwhm_L)

    return u + eta*(v-u)
  
def calc_eta(fwhm_G: float, fwhm_L: float) -> float:
  '''
  Calculate eta of pseudo voigt profile with fwhm_G, fwhm_L based on 
  Journal of Applied Crystallography. 33 (6): 1311–1316.

  Args:
    fwhm_G: full width at half maximum of gaussian part
    fwhm_L: full width at half maximum of lorenzian part
  Returns:
   maxing parameter eta
  '''
  f = fwhm_G**5+2.69269*fwhm_G**4*fwhm_L + \
    2.42843*fwhm_G**3*fwhm_L**2 + \
    4.47163*fwhm_G**2*fwhm_L**3 + \
    0.07842*fwhm_G*fwhm_L**4 + \
    fwhm_L**5
  f = f**(1/5)
  x = fwhm_L/f
  eta = 1.36603*x-0.47719*x**2+0.11116*x**3
  return eta



# fit osc
# fitting residual of time scan data with
# convolution of sum of damped oscillation and
# normalized gaussian distribution
# normalized cauchy distribution
# normalized pseudo voigt profile
# (Mixing parameter eta is fixed according to
#  Journal of Applied Crystallography. 33 (6): 1311–1316.)

import argparse
import numpy as np
from ..mathfun.A_matrix import make_A_matrix_dmp_osc, fact_anal_A
from ..mathfun import dmp_osc_conv, fact_anal_dmp_osc_conv
from .misc import set_bound_tau, read_data, contribution_table, plot_result
from lmfit import Parameters, fit_report, minimize

description = '''
fit osc: fitting residual of experimental time trace spectrum data with the convolution of the sum of damped oscilliation and irf function
There are three types of irf function (gaussian, cauchy, pseudo voigt)
It uses lmfit python package to fitting residual of time trace data and estimates error bound of irf parameter, lifetime constants and
period of vibrations. 
To calculate the contribution of each damped oscilliation component, it solve least linear square problem via scipy linalg lstsq module.
'''

epilog = '''
*Note

1. The number of tau, period and phase parameter should be same

2. The number of time zero should be same as the number of residual scan file to fit.

3. phase should be confined in [-pi/2, pi/2] (pi ~ 3.14)

4. If you set shape of irf to pseudo voigt (pv), then
   you should provide two full width at half maximum
   value for gaussian and cauchy parts, respectively.
'''

irf_help = '''
shape of instrument response functon
g: gaussian distribution
c: cauchy distribution
pv: pseudo voigt profile, linear combination of gaussian distribution and cauchy distribution 
    pv = eta*c+(1-eta)*g 
    the mixing parameter is fixed according to Journal of Applied Crystallography. 33 (6): 1311–1316. 
'''

fwhm_G_help = '''
full width at half maximum for gaussian shape
It would not be used when you set cauchy irf function
'''

fwhm_L_help = '''
full width at half maximum for cauchy shape
It would not be used when you did not set irf or use gaussian irf function
'''


def fit_osc():

    def residual(params, t, num_comp, irf, data=None, eps=None):
        if irf in ['g', 'c']:
            fwhm = params['fwhm']
        else:
            fwhm = np.array([params['fwhm_G'], params['fwhm_L']])
        tau = np.zeros(num_comp)
        T = np.zeros(num_comp)
        phase = np.zeros(num_comp)
        for i in range(num_comp):
            tau[i] = params[f'tau_{i+1}']
            T[i] = params[f'period_{i+1}']
            phase[i] = params[f'phi_{i+1}']
        chi = np.zeros((data.shape[0], data.shape[1]))
        for i in range(data.shape[1]):
            t0 = params[f't_0_{i+1}']
            A = make_A_matrix_dmp_osc(t-t0, fwhm, tau, T, phase, irf)
            c = fact_anal_A(A, data[:,i], eps[:, i])
            chi[:, i] = data[:, i] - (c@A)
        chi = chi.flatten()/eps.flatten()

        return chi

    tmp = argparse.RawTextHelpFormatter
    parser = argparse.ArgumentParser(formatter_class=tmp,
                                     description=description,
                                     epilog=epilog)
    parser.add_argument('--irf', default='g', choices=['g', 'c', 'pv'],
                        help=irf_help)
    parser.add_argument('--fwhm_G', type=float,
                        help=fwhm_G_help)
    parser.add_argument('--fwhm_L', type=float,
                        help=fwhm_L_help)
    parser.add_argument('prefix',
                        help='prefix of residual of time delay scan data to fit' +
                        'It will prefix_i.txt')
    parser.add_argument('-t0', '--time_zeros', type=float, nargs='+',
                        help='time zeros for each residual of tscan')
    parser.add_argument('-t0f', '--time_zeros_file',
                        help='filename for time zeros of each residual of tscan')
    parser.add_argument('--tau', type=float, nargs='+',
                        help='lifetime of each component')
    parser.add_argument('--period', type=float, nargs='+',
                        help='period of the vibration of each component')
    parser.add_argument('--phase', type=float, nargs='+',
    help='phase factor of each damped oscilliation component')
    parser.add_argument('--fix_irf', action='store_true',
    help='fix irf parameter (fwhm_G, fwhm_L) during fitting process')
    parser.add_argument('--fix_t0', action='store_true',
    help='fix time zero during fitting process')
    parser.add_argument('--slow', action='store_true',
    help='use slower but robust global optimization algorithm')
    parser.add_argument('-o', '--out', default=None,
                        help='prefix for output files')
    args = parser.parse_args()

    prefix = args.prefix
    if args.out is None:
        args.out = prefix
    out_prefix = args.out

    irf = args.irf
    if irf == 'g':
        if args.fwhm_G is None:
            print('You are using gaussian irf, so you should set fwhm_G!\n')
            return
        else:
            fwhm = args.fwhm_G
    elif irf == 'c':
        if args.fwhm_L is None:
            print('You are using cauchy/lorenzian irf,' +
                  'so you should set fwhm_L!\n')
            return
        else:
            fwhm = args.fwhm_L
    else:
        if (args.fwhm_G is None) or (args.fwhm_L is None):
            print('You are using pseudo voigt irf,' +
                  'so you should set both fwhm_G and fwhm_L!\n')
            return
        else:
            fwhm = 0.5346*args.fwhm_L + \
                np.sqrt(0.2166*args.fwhm_L**2+args.fwhm_G**2)


    tau = np.array(args.tau)
    T = np.array(args.period)
    phi = np.array(args.phase)
    num_comp = tau.size

    if (args.time_zeros is None) and (args.time_zeros_file is None):
        print('You should set either time_zeros or time_zeros_file!\n')
        return
    elif args.time_zeros is None:
        time_zeros = np.genfromtxt(args.time_zeros_file)
        num_scan = time_zeros.size
    else:
        time_zeros = np.array(args.time_zeros)
        num_scan = time_zeros.size

    t = np.genfromtxt(f'{prefix}_1.txt')[:, 0]
    num_data_pts = t.size
    data, eps = read_data(prefix, num_scan, num_data_pts, 10)

    print(f'fitting with {num_scan} data set!\n')
    fit_params = Parameters()
    if irf in ['g', 'c']:
        fit_params.add('fwhm', value=fwhm,
                       min=0.5*fwhm, max=2*fwhm, vary=(not args.fix_irf))
    elif irf == 'pv':
        fit_params.add('fwhm_G', value=args.fwhm_G,
                       min=0.5*args.fwhm_G, max=2*args.fwhm_G, vary=(not args.fix_irf))
        fit_params.add('fwhm_L', value=args.fwhm_L,
                       min=0.5*args.fwhm_L, max=2*args.fwhm_L, vary=(not args.fix_irf))
    for i in range(num_scan):
        fit_params.add(f't_0_{i+1}', value=time_zeros[i],
                       min=time_zeros[i]-2*fwhm,
                       max=time_zeros[i]+2*fwhm, vary=(not args.fix_t0))


    for i in range(num_comp):
        bd1 = set_bound_tau(tau[i])
        bd2 = set_bound_tau(T[i])
        fit_params.add(f'tau_{i+1}', value=tau[i], min=bd1[0], max=bd1[1])
        fit_params.add(f'period_{i+1}', value=T[i], min=bd2[0], max=bd2[1])
        fit_params.add(f'phi_{i+1}', value=phi[i], min=-np.pi/2, max=np.pi/2+0.1)

    # Second initial guess using global optimization algorithm
    if args.slow: 
        out = minimize(residual, fit_params, method='ampgo', calc_covar=False,
        args=(t, num_comp, irf),
        kws={'data': data, 'eps': eps})
    else:
        out = minimize(residual, fit_params, method='nelder', calc_covar=False,
        args=(t, num_comp, irf),
        kws={'data': data, 'eps': eps})

    # Then do Levenberg-Marquardt
    out = minimize(residual, out.params,
                   args=(t, num_comp),
                   kws={'data': data, 'eps': eps, 'irf': irf})

    chi2_ind = residual(out.params, t, num_comp, irf, data=data, eps=eps)
    chi2_ind = chi2_ind.reshape(data.shape)
    chi2_ind = np.sum(chi2_ind**2, axis=0)/(data.shape[0]-len(out.params))

    fit = np.zeros((data.shape[0], data.shape[1]+1))
    fit[:, 0] = t

    if irf in ['g', 'c']:
        fwhm_opt = out.params['fwhm']
    else:
        tmp_G = out.params['fwhm_G']
        tmp_L = out.params['fwhm_L']
        fwhm_opt = np.array([tmp_G, tmp_L])

    tau_opt = np.zeros(num_comp)
    T_opt = np.zeros(num_comp)
    phi_opt = np.zeros(num_comp)

    for j in range(num_comp):
        tau_opt[j] = out.params[f'tau_{j+1}']
        T_opt[j] = out.params[f'period_{j+1}']
        phi_opt[j] = out.params[f'phi_{j+1}']

    c = np.zeros((num_comp, num_scan))
    for i in range(num_scan):
        c[:, i] = fact_anal_dmp_osc_conv(t-out.params[f't_0_{i+1}'], fwhm_opt,
        tau_opt, T_opt, phi_opt, irf=irf, data=data[:, i], eps=eps[:, i])
        fit[:, i+1] = dmp_osc_conv(t-out.params[f't_0_{i+1}'], fwhm_opt,
        tau_opt, T_opt, phi_opt, c[:, i], irf=irf)
    
    contrib_table = contribution_table('tscan', 'Component Contribution',
    num_scan, num_comp, c)

    fit_content = fit_report(out) + '\n' + contrib_table

    print(fit_content)

    f = open(out_prefix+'_fit_report.txt', 'w')
    f.write(fit_content)
    f.close()

    np.savetxt(out_prefix+'_fit.txt', fit)
    np.savetxt(out_prefix+'_c.txt', c)

    plot_result('res tscan', num_scan, chi2_ind, data, eps, fit)

    return



'''
mathfun:
subpackage for the mathematical functions for TRXASprefitpack

:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3.
'''

from .broad import gen_theory_data
from .irf import gau_irf, cauchy_irf, pvoigt_irf, calc_eta
from .exp_conv_irf import exp_conv_gau, exp_conv_cauchy, exp_conv_pvoigt
from .exp_conv_irf import dmp_osc_conv_gau, dmp_osc_conv_cauchy, dmp_osc_conv_pvoigt
from .rate_eq import solve_model, solve_seq_model, solve_l_model, compute_model
from .rate_eq import compute_signal_gau, compute_signal_cauchy
from .rate_eq import compute_signal_pvoigt
from .exp_decay_fit import model_n_comp_conv, fact_anal_exp_conv
from .exp_decay_fit import rate_eq_conv, fact_anal_rate_eq_conv
from .exp_decay_fit import dmp_osc_conv, fact_anal_dmp_osc_conv


__all__ = ['gen_theory_data',
           'gau_irf', 'cauchy_irf', 'pvoigt_irf', 'calc_eta',
           'exp_conv_gau', 'exp_conv_cauchy', 'exp_conv_pvoigt',
           'dmp_osc_conv_gau', 'dmp_osc_conv_cauchy', 'dmp_osc_conv_pvoigt',
           'solve_model', 'solve_seq_model', 'solve_l_model', 'compute_model',
           'compute_signal_gau', 'compute_signal_cauchy', 'compute_signal_pvoigt', 
           'model_n_comp_conv', 'fact_anal_exp_conv',
           'rate_eq_conv', 'fact_anal_rate_eq_conv',
           'dmp_osc_conv', 'fact_anal_dmp_osc_conv']



# fit irf
# fitting irf function
# There are three irf shapes are avaliable
# normalized gaussian distribution
# normalized cauchy distribution
# normalized pseudo voigt profile
# (Mixing parameter eta is fixed according to
#  Journal of Applied Crystallography. 33 (6): 1311–1316.)

import argparse
import numpy as np
from ..mathfun.irf import gau_irf, cauchy_irf, pvoigt_irf, calc_eta
from .misc import read_data, plot_result
from lmfit import Parameters, fit_report, minimize

description = '''
fit irf: fitting shape of instrumental response function
There are three irf shapes are avaliable.
1. gaussian 2. cauchy(lorenzian) 3. pseudo voigt shape
It uses lmfit python package to fit experimentally measured instrumental response function to three model irf shape.
'''

irf_help = '''
shape of instrument response functon
g: gaussian distribution
c: cauchy distribution
pv: pseudo voigt profile, linear combination of gaussian distribution and cauchy distribution 
    pv = eta*c+(1-eta)*g 
    the mixing parameter is fixed according to Journal of Applied Crystallography. 33 (6): 1311–1316. 
'''

fwhm_G_help = '''
full width at half maximum for gaussian shape
It would not be used when you set cauchy irf function
'''

fwhm_L_help = '''
full width at half maximum for cauchy shape
It would not be used when you did not set irf or use gaussian irf function
'''

epilog = '''
*Note

1. The number of time zero parameter should be same as the
   number of scan to fit.

2. if you set shape of irf to pseudo voigt (pv), then
   you should provide two full width at half maximum
   value for gaussian and cauchy parts, respectively.

3. This script is only useful when one can directly measure
   instrumental response function.
'''


def fit_irf():

    def residual(params, t, irf, data=None, eps=None):
        if irf in ['g', 'c']:
            fwhm = params['fwhm']
        else:
            fwhm = np.array([params['fwhm_G'], params['fwhm_L']])
        chi = np.zeros((data.shape[0], data.shape[1]))
        for i in range(data.shape[1]):
            t0 = params[f't_0_{i+1}']
            if irf == 'g':
                chi[:, i] = data[:, i] - gau_irf(t-t0, fwhm)
            elif irf == 'c':
                chi[:, i] = data[:, i] - cauchy_irf(t-t0, fwhm)
            else:
                eta = calc_eta(fwhm[0], fwhm[1])
                chi[:, i] = data[:, i] - pvoigt_irf(t-t0, fwhm[0], fwhm[1], eta)
        chi = chi.flatten()/eps.flatten()

        return chi

    tmp = argparse.RawTextHelpFormatter
    parser = argparse.ArgumentParser(formatter_class=tmp,
                                     description=description,
                                     epilog=epilog)
    parser.add_argument('--irf', default='g', choices=['g', 'c', 'pv'],
                        help=irf_help)
    parser.add_argument('--fwhm_G', type=float,
                        help=fwhm_G_help)
    parser.add_argument('--fwhm_L', type=float,
                        help=fwhm_L_help)
    parser.add_argument('prefix',
                        help='prefix for tscan files ' +
                        'It will read prefix_i.txt')
    parser.add_argument('-t0', '--time_zeros', type=float, nargs='+',
                        help='time zeros for each tscan')
    parser.add_argument('-t0f', '--time_zeros_file',
                        help='filename for time zeros of each tscan')
    parser.add_argument('--slow', action='store_true',
    help='use slower but robust global optimization algorithm')
    parser.add_argument('-o', '--out', default=None,
                        help='prefix for output files')
    args = parser.parse_args()

    prefix = args.prefix
    if args.out is None:
        args.out = prefix
    out_prefix = args.out

    irf = args.irf
    if irf == 'g':
        if args.fwhm_G is None:
            print('You are using gaussian irf, so you should set fwhm_G!\n')
            return
        else:
            fwhm = args.fwhm_G
    elif irf == 'c':
        if args.fwhm_L is None:
            print('You are using cauchy/lorenzian irf,' +
                  'so you should set fwhm_L!\n')
            return
        else:
            fwhm = args.fwhm_L
    else:
        if (args.fwhm_G is None) or (args.fwhm_L is None):
            print('You are using pseudo voigt irf,' +
                  'so you should set both fwhm_G and fwhm_L!\n')
            return
        else:
            fwhm = 0.5346*args.fwhm_L + \
                np.sqrt(0.2166*args.fwhm_L**2+args.fwhm_G**2)

    if (args.time_zeros is None) and (args.time_zeros_file is None):
        print('You should set either time_zeros or time_zeros_file!\n')
        return
    elif args.time_zeros is None:
        time_zeros = np.genfromtxt(args.time_zeros_file)
        num_scan = time_zeros.size
    else:
        time_zeros = np.array(args.time_zeros)
        num_scan = time_zeros.size

    t = np.genfromtxt(f'{prefix}_1.txt')[:, 0]
    num_data_pts = t.size
    data, eps = read_data(prefix, num_scan, num_data_pts, 10)

    print(f'fitting with {num_scan} data set!\n')
    fit_params = Parameters()
    if irf in ['g', 'c']:
        fit_params.add('fwhm', value=fwhm,
                       min=0.5*fwhm, max=2*fwhm)
    elif irf == 'pv':
        fit_params.add('fwhm_G', value=args.fwhm_G,
                       min=0.5*args.fwhm_G, max=2*args.fwhm_G)
        fit_params.add('fwhm_L', value=args.fwhm_L,
                       min=0.5*args.fwhm_L, max=2*args.fwhm_L)
    for i in range(num_scan):
        fit_params.add(f't_0_{i+1}', value=time_zeros[i],
                       min=time_zeros[i]-2*fwhm,
                       max=time_zeros[i]+2*fwhm)

    # Second initial guess using global optimization algorithm
    if args.slow: 
        out = minimize(residual, fit_params, method='ampgo', calc_covar=False,
        args=(t, irf),
        kws={'data': data, 'eps': eps})
    else:
        out = minimize(residual, fit_params, method='nelder', calc_covar=False,
        args=(t, irf),
        kws={'data': data, 'eps': eps})

    # Then do Levenberg-Marquardt
    out = minimize(residual, out.params,
                   args=(t, irf),
                   kws={'data': data, 'eps': eps})

    chi2_ind = residual(out.params, t, irf, data=data, eps=eps)
    chi2_ind = chi2_ind.reshape(data.shape)
    chi2_ind = np.sum(chi2_ind**2, axis=0)/(data.shape[0]-len(out.params))

    fit = np.zeros((data.shape[0], data.shape[1]+1))
    fit[:, 0] = t

    for i in range(num_scan):
        if irf in ['g', 'c']:
            fwhm_out = out.params['fwhm']
        else:
            tmp_G = out.params['fwhm_G']
            tmp_L = out.params['fwhm_L']
            fwhm_out = np.array([tmp_G, tmp_L])
        
        if irf == 'g':
            fit[:, i+1] = gau_irf(t-out.params[f't_0_{i+1}'],
            fwhm_out)
        elif irf == 'c':
            fit[:, i+1] = cauchy_irf(t-out.params[f't_0_{i+1}'],
            fwhm_out)
        else:
            f_out = fwhm_out[0]**5+2.69269*fwhm_out[0]**4*fwhm_out[1]+2.42843*fwhm_out[0]**3*fwhm_out[1]**2 + \
                    4.47163*fwhm_out[0]**2*fwhm_out[1]**3 + 0.07842*fwhm_out[0]*fwhm_out[1]**4 + fwhm_out[1]**5
            f_out = f_out**(1/5)
            eta_out = 1.36603*(fwhm_out[1]/f_out)-0.47719*(fwhm_out[1]/f_out)**2+0.11116*(fwhm_out[1]/f_out)**3
            fit[:, i+1] = pvoigt_irf(t-out.params[f't_0_{i+1}'], fwhm_out[0], fwhm_out[1], eta_out)
    
    f = open(out_prefix+'_fit_report.txt', 'w')
    f.write(fit_report(out))
    f.close()

    np.savetxt(out_prefix+'_fit.txt', fit)

    print(fit_report(out))

    plot_result('irf_scan', num_scan, chi2_ind, data, eps, fit)

    return



# fit eq py
# Wrapper script for fit_eq()
# Date: 2022. 6. 21.
# Author: pistack
# Email: pistatex@yonsei.ac.kr

import os
import sys

path = os.path.dirname(os.path.realpath(__file__))
sys.path.append(path+"/../src/")

from TRXASprefitpack.tools import fit_eq

if __name__ == "__main__":
    fit_eq()



'''
exp_conv_irf:
submodule for the mathematical functions for
exponential decay convolved with irf

:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3.
'''

from typing import Union
import numpy as np
from scipy.special import erfc, erfcx, wofz, exp1


def exp_conv_gau(t: Union[float, np.ndarray], fwhm: float,
                 k: float) -> Union[float, np.ndarray]:

    '''
    Compute exponential function convolved with normalized gaussian
    distribution

    Args:
      t: time
      fwhm: full width at half maximum of gaussian distribution
      k: rate constant (inverse of life time)

    Returns:
     Convolution of normalized gaussian distribution and exponential
     decay :math:`(\\exp(-kt))`.
    '''

    sigma = fwhm/(2*np.sqrt(2*np.log(2)))
    ksigma = k*sigma
    z = ksigma - t/sigma

    if not isinstance(t, np.ndarray):
        if z > 0:
            ans = 1/2*np.exp(-(t/sigma)**2/2) * \
                erfcx(z/np.sqrt(2))
        else:
            ans = 1/2*np.exp(ksigma*z-ksigma**2/2) * \
                erfc(z/np.sqrt(2))
    else:
        mask = z > 0
        inv_mask = np.invert(mask)
        ans = np.zeros(t.shape[0], dtype=np.float64)
        ans[mask] = 1/2*np.exp(-(t[mask]/sigma)**2/2) * \
            erfcx(z[mask]/np.sqrt(2))
        ans[inv_mask] = 1/2*np.exp(ksigma*z[inv_mask]-ksigma**2/2) * \
            erfc(z[inv_mask]/np.sqrt(2))

    return ans

def exp1x_asymp(z: Union[complex, np.ndarray]) -> Union[complex, np.ndarray]:
    return z*(1+z*(1+2*z*(1+3*z*(1+4*z*(1+5*z*(1+6*z*(1+7*z*(1+8*z*(1+9*z*(1+10*z))))))))))

def exp_conv_cauchy(t: Union[float, np.ndarray],
                    fwhm: float,
                    k: float) -> Union[float, np.ndarray]:

    '''
    Compute exponential function convolved with normalized cauchy
    distribution

    Args:
       t: time
       fwhm: full width at half maximum of cauchy distribution
       k: rate constant (inverse of life time)

    Returns:
      Convolution of normalized cauchy distribution and
      exponential decay :math:`(\\exp(-kt))`.
    '''

    if k == 0:
        ans = 0.5+1/np.pi*np.arctan(2*t/fwhm)
    else:
        kgamma = k*fwhm/2
        ikgamma = complex(0, kgamma)
        kt = k*t
        if not isinstance(t, np.ndarray):
            if np.abs(kt) < 200:
                ans = complex(np.cos(kgamma), -np.sin(kgamma))*exp1(-kt-ikgamma)
                ans = np.exp(-kt)*ans.imag/np.pi
            else:
                inv_z = 1/(kt+ikgamma)
                ans = exp1x_asymp(inv_z)
                ans = -ans.imag/np.pi
        else:
            mask = np.abs(kt) < 200
            inv_mask = np.invert(mask)

            ans = np.zeros(t.shape[0], dtype=np.float64)
            inv_z = 1/(kt[inv_mask]+ikgamma)

            # abs(kt) < 200
            ans1 = complex(np.cos(kgamma), -np.sin(kgamma))*exp1(-kt[mask]-ikgamma)
            ans[mask] = np.exp(-kt[mask])*ans1.imag/np.pi

            # abs(kt) > 200, use asymptotic series
            ans2 = exp1x_asymp(inv_z)
            ans[inv_mask] = -ans2.imag/np.pi
    return ans


def exp_conv_pvoigt(t: Union[float, np.ndarray],
                    fwhm_G: float,
                    fwhm_L: float,
                    eta: float,
                    k: float) -> Union[float, np.ndarray]:

    '''
    Compute exponential function convolved with normalized pseudo
    voigt profile (i.e. linear combination of normalized gaussian and
    cauchy distribution)

    :math:`\\eta C(\\mathrm{fwhm}_L, t) + (1-\\eta)G(\\mathrm{fwhm}_G, t)`

    Args:
       t: time
       fwhm_G: full width at half maximum of gaussian part of
               pseudo voigt profile
       fwhm_L: full width at half maximum of cauchy part of
               pseudo voigt profile
       eta: mixing parameter
       k: rate constant (inverse of life time)

    Returns:
       Convolution of normalized pseudo voigt profile and
       exponential decay :math:`(\\exp(-kt))`.
    '''

    u = exp_conv_gau(t, fwhm_G, k)
    v = exp_conv_cauchy(t, fwhm_L, k)

    return u + eta*(v-u)

def dmp_osc_conv_gau(t: Union[float, np.ndarray], fwhm: float,
k: float, T: float, phase: float) -> Union[float, np.ndarray]:

    '''
    Compute damped oscillation convolved with normalized gaussian
    distribution

    Args:
      t: time
      fwhm: full width at half maximum of gaussian distribution
      k: damping constant (inverse of life time)
      T: period of vibration 
      phase: phase factor

    Returns:
     Convolution of normalized gaussian distribution and 
     damped oscillation :math:`(\\exp(-kt)cos(2\\pi t/T+phase))`.
    '''

    sigma = fwhm/(2*np.sqrt(2*np.log(2)))
    sigmak_cplx = complex(k*sigma, -2*np.pi*sigma/T)
    z = sigmak_cplx - t/sigma

    if not isinstance(t, np.ndarray):
        if z.real > 0:
            ans = 1/2*np.exp(-(t/sigma)**2/2) * \
                wofz(z/complex(0, -np.sqrt(2)))
        else:
            ans = np.exp(sigmak_cplx*z-sigmak_cplx**2/2) - \
                1/2*np.exp(-(t/sigma)**2/2)*wofz(z/complex(0, np.sqrt(2))) 
    else:
        mask = z.real > 0
        inv_mask = np.invert(mask)
        ans = np.zeros(t.shape[0], dtype=np.complex128)
        ans[mask] = 1/2*np.exp(-(t[mask]/sigma)**2/2) * \
            wofz(z[mask]/complex(0, -np.sqrt(2)))
        ans[inv_mask] = np.exp(sigmak_cplx*z[inv_mask]-sigmak_cplx**2/2) - \
            1/2*np.exp(-(t[inv_mask]/sigma)**2/2)*wofz(z[inv_mask]/complex(0, np.sqrt(2)))
    
    ans = (complex(np.cos(phase), np.sin(phase)))*ans

    return ans.real

def dmp_osc_conv_cauchy(t: Union[float, np.ndarray], fwhm: float,
k: float, T: float, phase: float) -> Union[float, np.ndarray]:

    '''
    Compute damped oscillation convolved with normalized cauchy
    distribution

    Args:
      t: time
      fwhm: full width at half maximum of cauchy distribution
      k: damping constant (inverse of life time)
      T: period of vibration 
      phase: phase factor

    Returns:
     Convolution of normalized cauchy distribution and 
     damped oscillation :math:`(\\exp(-kt)cos(2\\pi t/T+phase))`.
    '''

    gamma = fwhm/2; omega = 2*np.pi/T 
    z1 = (-k*t-omega*gamma) + complex(0,1)*(-k*gamma+omega*t)
    z2 = (-k*t+omega*gamma) + complex(0,-1)*(k*gamma+omega*t)
    
    if not isinstance(t, np.ndarray):
        if np.abs(z1.real) < 200:
            ans1 = np.exp(z1.real)*complex(np.cos(z1.imag+phase), np.sin(z1.imag+phase))*\
                (exp1(z1)+complex(0, 2*np.pi*np.heaviside(z1.imag, 1)))
        else:
            ans1 = exp1x_asymp(-1/z1)*complex(-np.cos(phase), -np.sin(phase))
        if np.abs(z2.real) < 200:
            ans2 =  np.exp(z2.real)*complex(np.cos(z2.imag-phase), np.sin(z1.imag-phase))*\
                exp1(z2)
        else:
            ans2 = exp1x_asymp(-1/z2)*complex(-np.cos(phase), np.sin(phase))
    
    else:
        ans1 = np.zeros_like(t, dtype=np.complex128)
        ans2 = np.zeros_like(t, dtype=np.complex128)
        mask1 = np.abs(z1.real) < 200; inv_mask1 = np.invert(mask1)
        mask2 = np.abs(z2.real) < 200; inv_mask2 = np.invert(mask2)

        ans1[mask1] = np.exp(z1[mask1].real)*(np.cos(z1[mask1].imag+phase)+complex(0, 1)*np.sin(z1[mask1].imag+phase))*\
            (exp1(z1[mask1])+complex(0, 2*np.pi)*np.heaviside(z1[mask1].imag, 1))
        ans1[inv_mask1] = exp1x_asymp(-1/z1[inv_mask1])*complex(-np.cos(phase), -np.sin(phase))

        ans2[mask2] = np.exp(z2[mask2].real)*(np.cos(z2[mask2].imag-phase)+complex(0, 1)*np.sin(z2[mask2].imag-phase))*\
            exp1(z2[mask2])
        ans2[inv_mask2] = exp1x_asymp(-1/z2[inv_mask2])*complex(-np.cos(phase), np.sin(phase))
        
    ans = (ans1.imag + ans2.imag)/(2*np.pi)

    return ans

def dmp_osc_conv_pvoigt(t: Union[float, np.ndarray], fwhm_G: float, fwhm_L: float, eta: float,
k: float, T: float, phase: float) -> Union[float, np.ndarray]:

    '''
    Compute damped oscillation convolved with normalized pseudo
    voigt profile (i.e. linear combination of normalized gaussian and
    cauchy distribution)

    :math:`\\eta C(\\mathrm{fwhm}_L, t) + (1-\\eta)G(\\mathrm{fwhm}_G, t)`

    Args:
       t: time
       fwhm_G: full width at half maximum of gaussian part of
               pseudo voigt profile
       fwhm_L: full width at half maximum of cauchy part of
               pseudo voigt profile
       eta: mixing parameter
       k: damping constant (inverse of life time)
       T: period of vibration 
       phase: phase factor

    Returns:
     Convolution of normalized pseudo voigt profile and
     damped oscillation :math:`(\\exp(-kt)cos(2\\pi t/T+phase))`.
    '''

    u = dmp_osc_conv_gau(t, fwhm_G, k, T, phase)
    v = dmp_osc_conv_cauchy(t, fwhm_L, k, T, phase)

    return u + eta*(v-u)





# fit static
# fitting static spectrum
# with tddft calc peaks convolved by
# option
# v: voigt profile
# g: gaussian
# l: lorenzian
# to correct baseline I use linear line as baseline

import argparse
import numpy as np
import scipy.linalg as LA
from ..mathfun import gen_theory_data
from .misc import read_data, plot_result
from lmfit import Parameters, fit_report, minimize


def fit_static():

    def magick(e, thy_data, no_base, data=None, eps=None):
        if no_base:
            A = thy_data/eps
            y = data/eps
            c, _, _, _ = LA.lstsq(A.reshape(A.size, 1), y)
            c = np.hstack((c,np.zeros(2)))
        else:
            A = np.ones((e.shape[0], 3))
            A[:, 0] = thy_data
            A[:, 1] = e
            A[:, 0] = A[:, 0]/eps
            A[:, 1] = A[:, 1]/eps
            A[:, 2] = A[:, 2]/eps
            y = data/eps
            c, _, _, _ = LA.lstsq(A, y)
        return c

    def residual(params, e, peaks, policy, no_base, data=None, eps=None):
        fwhm_G = params['fwhm_G']
        fwhm_L = params['fwhm_L']
        peak_factor = params['peak_factor']
        chi = np.zeros(data.shape)
        thy_data = gen_theory_data(e, peaks, 1, fwhm_G, fwhm_L, peak_factor, policy)
        for i in range(data.shape[1]):
            c = magick(e, thy_data, no_base,
                       data=data[:, i], eps=eps[:, i])
            chi[:, i] = data[:, i] - (c[0]*thy_data+c[1]*e+c[2])
        chi = chi.flatten()/eps.flatten()
        return chi

    description = '''
fit static: fitting static spectrum with theoretically calculated line spectrum
broadened by spectral line shape
v: voigt profile,
g: gaussian,
l: lorenzian,
It uses lmfit python package to fit experimental spectrum and estimates the error bound of
broadening and peak parameter
Moreover, it uses linear baseline to correct baseline feature of experimental spectrum
'''

    epilog = ''' 
*Note
energy unit for measured static spectrum and theoretically calculated spectrum should be same
'''
    tmp = argparse.RawDescriptionHelpFormatter
    parse = argparse.ArgumentParser(formatter_class=tmp,
                                    description=description,
                                    epilog=epilog)
    parse.add_argument('-ls', '--line_shape',
                       default='v', choices=['v', 'g', 'l'],
                       help="line shape of spectrum"+'\n' +
                       "v: voigt profile" + '\n' +
                       "g: gaussian shape" + '\n' +
                       "l: lorenzian shape")
    parse.add_argument('--fwhm_G', type=float,
                        help='full width at half maximum for gaussian shape ' +
                        'It would be not used when you set lorenzian line shape')
    parse.add_argument('--fwhm_L', type=float,
                        help='full width at half maximum for lorenzian shape ' +
                        'It would be not used when you use gaussian line shape')
    parse.add_argument('--no_base', action='store_true',
    help ='Do not include linear base line during fitting process')
    parse.add_argument('--scale_energy', action='store_true',
    help='Scaling the energy of peak instead of shifting to match experimental spectrum')
    parse.add_argument('prefix',
                       help='prefix for experimental spectrum files' +
                       '\n' + 'It will read prefix_i.txt files')
    parse.add_argument('num_scan', type=int,
                       help='the number of static peak scan files')
    parse.add_argument('peak_file',
                       help='filename for theoretical line shape spectrum')
    parse.add_argument('peak_factor', type=float,
    help='parameter to match descrepency between thoretical spectrum and experimental spectrum')
    parse.add_argument('-o', '--out', help='prefix for output files')
    parse.add_argument('--fix_fwhm_G', action='store_true',
    help='fix gaussian fwhm value')
    parse.add_argument('--fix_fwhm_L', action='store_true',
    help='fix lorenzian fwhm value')
    parse.add_argument('--slow', action='store_true',
    help='use slower but robust global optimization algorithm')

    args = parse.parse_args()

    prefix = args.prefix
    option = args.line_shape
    num_scan = args.num_scan
    peak_file = args.peak_file
    peak_factor = args.peak_factor

    if args.out is None:
        out_prefix = prefix
    else:
        out_prefix = args.out

    if option == 'g':
        if args.fwhm_G is None:
            print("Please set fwhm_G of gaussian line shape")
            return
        else:
            fwhm = args.fwhm_G
    elif option == 'l':
        if args.fwhm_L is None:
            print("Please set fwhm_L of lorenzian line shape")
            return
        else:
            fwhm = args.fwhm_L
    else:
        if (args.fwhm_G is None) or (args.fwhm_L is None):
            print("Please set both fwhm_G and fwhm_L for Voigt line shape")
            return
        else:
            fwhm_lst = [args.fwhm_G, args.fwhm_L] 
    
    if args.scale_energy:
        policy = 'scale'
    else:
        policy = 'shift'

    no_base = args.no_base 

    e = np.genfromtxt(f'{prefix}_1.txt')[:, 0]
    num_escan_pts = e.shape[0]
    data, eps = read_data(prefix, num_scan, num_escan_pts, 1000)

    peaks = np.genfromtxt(peak_file)

    fit_params = Parameters()
    if option == 'v':
        fit_params.add('fwhm_G', value=fwhm_lst[0], min=fwhm_lst[0]/2, max=2*fwhm_lst[0],
        vary=(not args.fix_fwhm_G))
        fit_params.add('fwhm_L', value=fwhm_lst[1], min=fwhm_lst[1]/2, max=2*fwhm_lst[1],
        vary=(not args.fix_fwhm_L))
    elif option == 'g':
        fit_params.add('fwhm_G', value=fwhm, min=fwhm/2, max=2*fwhm, vary=(not args.fix_fwhm_G))
        fit_params.add('fwhm_L', value=0, vary=False)
    else:
        fit_params.add('fwhm_G', value=0, vary=False)
        fit_params.add('fwhm_L', value=fwhm, min=fwhm/2, max=2*fwhm, vary=(not args.fix_fwhm_L))

    fit_params.add('peak_factor', value=peak_factor,
                   min=peak_factor/2,
                   max=2*peak_factor)

    # First, Nelder-Mead (if slow use ampgo instead)
    if args.slow:
        out = minimize(residual, fit_params, method='ampgo', calc_covar=False,
        args=(e, peaks, policy, no_base),
        kws={'data': data, 'eps': eps})
    else:
        out = minimize(residual, fit_params, method='Nelder', calc_covar=False,
        args=(e, peaks, policy, no_base),
        kws={'data': data, 'eps': eps})
    # Then do Levenberg-Marquardt
    out = minimize(residual, out.params,
                   args=(e, peaks, policy, no_base),
                   kws={'data': data, 'eps': eps})

    print(fit_report(out))
    chi2_ind = residual(out.params, e, peaks, policy, no_base, data=data, eps=eps)
    chi2_ind = chi2_ind.reshape(data.shape)
    chi2_ind = np.sum(chi2_ind**2, axis=0)/(data.shape[0]-6)

    fwhm_G = out.params['fwhm_G']
    fwhm_L = out.params['fwhm_L']
    peak_factor = out.params['peak_factor']
    base = np.zeros((data.shape[0], data.shape[1]+1))
    fit = np.zeros((data.shape[0], data.shape[1]+1))
    base[:, 0] = e
    fit[:, 0] = e
    thy_data_opt = gen_theory_data(e, peaks, 1, fwhm_G, fwhm_L, peak_factor, policy)
    A = np.zeros(num_scan)
    for i in range(num_scan):
        c = magick(e, thy_data_opt, no_base, data=data[:, i],
                   eps=eps[:, i])
        base[:, i+1] = c[1]*e+c[2]
        fit[:, i+1] = c[0]*thy_data_opt+c[1]*e+c[2]
        A[i] = c[0]

    f = open(out_prefix+'_fit_report.txt', 'w')
    f.write(fit_report(out))
    f.close()

    np.savetxt(out_prefix+'_base.txt', base)
    np.savetxt(out_prefix+'_fit.txt', fit)
    np.savetxt(out_prefix+'_A.txt', A)

    reduced_fit = np.zeros_like(fit)
    reduced_fit[:, 0] = fit[:, 0]
    reduced_fit[:, 1:] = fit[:, 1:] - base[:, 1:]

    plot_result('static', num_scan, chi2_ind, data-base[:, 1:], eps, reduced_fit)

    return



'''
A_matrix:
submodule for evaluation of A_matrix

:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3.
'''
from typing import Union, Optional
import numpy as np
from scipy.linalg import lstsq
from .irf import calc_eta
from .exp_conv_irf import exp_conv_gau, exp_conv_cauchy
from .exp_conv_irf import dmp_osc_conv_gau, dmp_osc_conv_cauchy


def make_A_matrix(t: np.ndarray, k: np.ndarray) -> np.ndarray:

    A = np.zeros((k.size, t.size))
    for i in range(k.size):
        A[i, :] = np.exp(-k[i]*t)
    A = A*np.heaviside(t, 1)
    A[np.isnan(A)] = 0

    return A


def make_A_matrix_gau(t: np.ndarray, fwhm: float,
                      k: np.ndarray) -> np.ndarray:

    A = np.zeros((k.size, t.size))
    for i in range(k.size):
        A[i, :] = exp_conv_gau(t, fwhm, k[i])

    return A


def make_A_matrix_cauchy(t: np.ndarray, fwhm: float,
                         k: np.ndarray) -> np.ndarray:

    A = np.zeros((k.size, t.size))
    for i in range(k.size):
        A[i, :] = exp_conv_cauchy(t, fwhm, k[i])

    return A


def make_A_matrix_pvoigt(t: np.ndarray,
                         fwhm_G: float,
                         fwhm_L: float,
                         eta: float,
                         k: np.ndarray) -> np.ndarray:
    
    u = make_A_matrix_gau(t, fwhm_G, k)
    v = make_A_matrix_cauchy(t, fwhm_L, k)

    return u + eta*(v-u)

def make_A_matrix_exp(t: np.ndarray,
                      fwhm: Union[float, np.ndarray],
                      tau: np.ndarray,
                      base: Optional[bool] = True,
                      irf: Optional[str] = 'g',
                      eta: Optional[float] = None
                      ) -> np.ndarray:

    k = 1/tau
    if base:
        k = np.concatenate((k, np.array([0])))

    if irf == 'g':
        A = make_A_matrix_gau(t, fwhm, k)
    elif irf == 'c':
        A = make_A_matrix_cauchy(t, fwhm, k)
    elif irf == 'pv':
        if eta is None:
            eta = calc_eta(fwhm[0], fwhm[1])
        A = make_A_matrix_pvoigt(t, fwhm[0], fwhm[1],
                                 eta, k)

    return A

def make_A_matrix_gau_osc(t: np.ndarray, fwhm: float,
k: np.ndarray, T: np.ndarray, phase: np.ndarray) -> np.ndarray:

    A = np.zeros((k.size, t.size))
    for i in range(k.size):
        A[i, :] = dmp_osc_conv_gau(t, fwhm, k[i], T[i], phase[i])

    return A

def make_A_matrix_cauchy_osc(t: np.ndarray, fwhm: float,
k: np.ndarray, T: np.ndarray, phase: np.ndarray) -> np.ndarray:

    A = np.zeros((k.size, t.size))
    for i in range(k.size):
        A[i, :] = dmp_osc_conv_cauchy(t, fwhm, k[i], T[i], phase[i])

    return A

def make_A_matrix_pvoigt_osc(t: np.ndarray, fwhm_G: float, fwhm_L: float, eta: float,
k: np.ndarray, T: np.ndarray, phase: np.ndarray) -> np.ndarray:
    
    u = make_A_matrix_gau_osc(t, fwhm_G, k, T, phase)
    v = make_A_matrix_cauchy_osc(t, fwhm_L, k, T, phase)

    return u + eta*(v-u)

def make_A_matrix_dmp_osc(t: np.ndarray, fwhm: Union[float, np.ndarray],
                      tau: np.ndarray,
                      T: np.ndarray,
                      phase: np.ndarray,
                      irf: Optional[str] = 'g',
                      eta: Optional[float] = None
                      ) -> np.ndarray:


    if irf == 'g':
        A = make_A_matrix_gau_osc(t, fwhm, 1/tau, T, phase)
    elif irf == 'c':
        A = make_A_matrix_cauchy_osc(t, fwhm, 1/tau, T, phase)
    elif irf == 'pv':
        if eta is None:
            eta = calc_eta(fwhm[0], fwhm[1])
        A = make_A_matrix_pvoigt_osc(t, fwhm[0], fwhm[1], eta, 1/tau, T, phase)

    return A

def fact_anal_A(A: np.ndarray, data: Optional[np.ndarray] = None, eps: Optional[np.ndarray] = None) -> np.ndarray:
    B = np.copy(A)
    y = np.copy(data)
    if eps is None:
        eps = np.ones_like(data)
    
    B = np.einsum('j,ij->ij', 1/eps, B)
    y = y/eps
    
    c, _, _, _ = lstsq(B.T, y)
    return c



# misc
# submodule for miscellaneous function of
# tools subpackage

from typing import Tuple, Union, Optional
import numpy as np
import matplotlib.pyplot as plt

def set_bound_tau(tau: float):
    '''
    Setting bound for lifetime constant

    Args:
      tau: lifetime constant

    Returns:
     list of upper bound and lower bound of tau
    '''
    bound = [tau/2, 1]
    if 0.1 < tau <= 1:
        bound = [0.05, 10]
    elif 1 < tau <= 10:
        bound = [0.5, 50]
    elif 10 < tau <= 100:
        bound = [5, 500]
    elif 100 < tau <= 1000:
        bound = [50, 2000]
    elif 1000 < tau <= 5000:
        bound = [500, 10000]
    elif 5000 < tau <= 50000:
        bound = [2500, 100000]
    elif 50000 < tau <= 500000:
        bound = [25000, 1000000]
    elif 500000 < tau <= 1000000:
        bound = [250000, 2000000]
    elif 1000000 < tau:
        bound = [tau/4, 4*tau]
    return bound

def read_data(prefix: str, num_scan: int, num_data_pts: int, default_SN: float) -> Tuple[np.ndarray, np.ndarray]:
    '''
    Read data from prefix_i.txt (1 <= i <= num_scan)

    Args:
     prefix: prefix of scan_file
     num_scan: the number of scan file to read
     num_data_pts: the number of data points per each scan file
     default_SN: Default Signal/Noise
    
    Return:
     Tuple of data of eps
    '''
    data = np.zeros((num_data_pts, num_scan))
    eps = np.zeros((num_data_pts, num_scan))

    for i in range(num_scan):
        A = np.genfromtxt(f'{prefix}_{i+1}.txt')
        data[:, i] = A[:, 1]
        if A.shape[1] == 2:
            eps[:, i] = np.max(np.abs(data[:, i]))*np.ones(num_data_pts)/default_SN
        else:
            eps[:, i] = A[:, 2]
    return data, eps

def plot_result(scan_name: str, num_scan: int, chi2_ind: Union[list, np.ndarray],
data: np.ndarray, eps: np.ndarray, fit: np.ndarray, res: Optional[np.ndarray] = None):
    '''
    Plot fitting result

    Args:
     scan_name: name of scan
     num_scan: the number of scans
     chi2_ind: list or array of the chi squared value of each indivisual scan
     data: exprimental data
     eps: error or data quality of experimental data
     fit: fitting result
     res: residual of fitting (data-fit)

    Note:
     1. the first column of fit array should be time range
     2. data array should not contain time range
    '''
    t = fit[:, 0]
    if res is not None:
        for i in range(num_scan):
            fig = plt.figure(i+1)
            title = f'Chi squared: {chi2_ind[i]:.2f}'
            plt.suptitle(title)
            sub1 = fig.add_subplot(211)
            sub1.errorbar(t, data[:, i],
            eps[:, i], marker='o', mfc='none',
            label=f'expt {scan_name} {i+1}',
            linestyle='none')
            sub1.plot(t, fit[:, i+1],
            label=f'fit {scan_name} {i+1}')
            sub1.legend()
            sub2 = fig.add_subplot(212)
            sub2.errorbar(t, res[:, i+1],
            eps[:, i], marker='o', mfc='none',
            label=f'{scan_name} res {i+1}',
            linestyle='none')
            sub2.legend()

    else:
        for i in range(num_scan):
            plt.figure(i+1)
            title = f'Chi squared: {chi2_ind[i]:.2f}'
            plt.title(title)
            plt.errorbar(t, data[:, i],
            eps[:, i], marker='o', mfc='none',
            label=f'expt {scan_name} {i+1}',
            linestyle='none')
            plt.plot(t, fit[:, i+1],
            label=f'fit {scan_name} {i+1}')
            plt.legend()
    plt.show()
    return

def contribution_table(scan_name: str, table_title: str, num_scan: int, num_comp: int, coeff: np.ndarray) -> str:
    '''
    Draw contribution table (row: num_comp, col: num_scan)

    Args:
     scan_name: name of scan
     table_title: title of table
     num_scan: the number of scan
     num_comp: the number of component
     coeff: coefficient matrix
     
    Return:
     contribution table
    '''
    # calculate contribution
    coeff_abs = np.abs(coeff)
    coeff_sum = np.sum(coeff_abs, axis=0)
    coeff_contrib = np.zeros_like(coeff)
    for i in range(num_scan):
        coeff_contrib[:, i] = coeff[:, i]/coeff_sum[i]*100

    # draw table
    cont_table = '    '
    for i in range(num_scan):
        cont_table = cont_table + f'{scan_name} {i+1} |'
    cont_table = cont_table + '\n'
    for i in range(num_comp):
        cont_table = cont_table + '    '
        for j in range(num_scan):
            cont_table = cont_table + f'{coeff_contrib[i, j]:.2f} % |'
        cont_table = cont_table + '\n'
    
    cont_table = f'[[{table_title}]]' + '\n' + cont_table
    return cont_table

def parse_matrix(mat_str: np.ndarray, tau: np.ndarray) -> np.ndarray:
    '''
    Parse user supplied rate equation matrix

    Args:
     mat_str: user supplied rate equation (lower triangular matrix)
     tau: life time constants (inverse of rate constant)
    
    Return:
     parsed rate equation matrix.
    
    Note:
     Every entry in the rate equation matrix should be
     '0', '1*ki', '-x.xxx*ki', 'x.xxx*ki' or '-(x.xxx*ki+y.yyy*kj+...)'
    '''

    L = np.zeros_like(mat_str, dtype=float)
    mask = (mat_str != '0')
    red_mat_str = mat_str[mask]
    red_L = np.zeros_like(red_mat_str, dtype=float)

    for i in range(red_mat_str.size):
        tmp = red_mat_str[i]
        if '-' in tmp:
            tmp = tmp.strip('-')
            tmp = tmp.strip('('); tmp = tmp.strip(')')
            k_lst = tmp.split('+')
            for k in k_lst:
                k_pair = k.split('*')
                red_L[i] = red_L[i] - float(k_pair[0])/tau[int(k_pair[1][1:])-1]
        else:
            tmp_pair = tmp.split('*')
            red_L[i] = float(tmp_pair[0])/tau[int(tmp_pair[1][1:])-1]
    
    L[mask] = red_L
    
    return L



'''
broad:
submodule for broading theoritical spectrum

:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3
'''

from typing import Optional
import numpy as np
from scipy.special import voigt_profile


def gen_theory_data(e: np.ndarray,
                    peaks: np.ndarray,
                    A: float,
                    fwhm_G: float,
                    fwhm_L: float,
                    peak_factor: float,
                    policy: Optional[str] = 'shift') -> np.ndarray:

    '''
    voigt broadening theoretically calculated lineshape spectrum

    Args:
        e: energy 
        A: scaling parameter
        fwhm_G: full width at half maximum of gaussian shape (unit: same as energy)
        fwhm_L: full width at half maximum of lorenzian shape (unit: same as energy)
        peak_factor: Peak factor, its behavior depends on policy.
        policy: Policy to match discrepency between experimental data and theoretical
                spectrum.

                * 'shift' : Default option, shift peak position by peak_factor
                * 'scale' : scale peak position by peak_factor

    Returns:
      numpy ndarray of voigt broadened theoritical lineshape spectrum
    '''

    sigma = fwhm_G/(2*np.sqrt(2*np.log(2)))
    gamma = fwhm_L/2

    num_e = e.size
    num_peaks = peaks.shape[0]
    v_matrix = np.zeros((num_e, num_peaks))
    peak_copy = np.copy(peaks[:, 0])
    if policy == 'shift':
      peak_copy = peak_copy - peak_factor
    else:
      peak_copy = peak_factor*peak_copy 
    for i in range(num_peaks):
        v_matrix[:, i] = voigt_profile(e-peak_copy[i], sigma, gamma)

    broadened_theory = A * v_matrix @ peaks[:, 1].reshape((num_peaks, 1))
    broadened_theory = broadened_theory.flatten()

    return broadened_theory



import argparse
import numpy as np
from ..mathfun import gen_theory_data


def broadening():
    description = '''
    broadening: generates voigt broadened theoritical calc spectrum
    '''

    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('peak',
                        help='filename for calculated line shape spectrum')
    parser.add_argument('e_min', type=float,
                        help='minimum energy')
    parser.add_argument('e_max', type=float,
                        help='maximum energy')
    parser.add_argument('e_step', type=float,
    help='energy step')
    parser.add_argument('A', type=float,
                        help='scale factor')
    parser.add_argument('fwhm_G', type=float,
                        help='Full Width at Half Maximum of gaussian shape')
    parser.add_argument('fwhm_L', type=float,
                        help='Full Width at Half Maximum of lorenzian shape')
    parser.add_argument('peak_factor', type=float,
    help='parameter to match descrepency between thoretical spectrum and experimental spectrum')
    parser.add_argument('--scale_energy', action='store_true',
    help='Scaling the energy of peak instead of shifting to match experimental spectrum')
    parser.add_argument('-o', '--out', help='prefix for output files')
    args = parser.parse_args()

    peak = np.genfromtxt(args.peak)
    if args.out is None:
        out = args.prefix
    else:
        out = args.out
    e_min = args.e_min
    e_max = args.e_max
    e_step = args.e_step
    A = args.A
    fwhm_G = args.fwhm_G
    fwhm_L = args.fwhm_L
    peak_factor = args.peak_factor
    if args.scale_energy:
        policy = 'scale'
    else:
        policy = 'shift' 
    e = np.linspace(e_min, e_max, int((e_max-e_min)/e_step)+1)

    broadened_thy = gen_theory_data(e, peak, A, fwhm_G, fwhm_L, peak_factor, policy)

    rescaled_stk = peak
    if policy == 'scale':
        rescaled_stk[:, 0] = rescaled_stk[:, 0]*peak_factor
    else:
        rescaled_stk[:, 0] = rescaled_stk[:, 0] - peak_factor
    rescaled_stk[:, 1] = A*rescaled_stk[:, 1]
    spec_thy = np.vstack((e, broadened_thy))

    np.savetxt(f'{out}_thy_stk.txt', rescaled_stk, fmt=['%.5e', '%.8e'])
    np.savetxt(f'{out}_thy.txt', spec_thy.T, fmt=['%.5e', '%.8e'])

    return



'''
tools:
subpackage for TRXASprefitpack utilities

:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3
'''

from .fit_static import fit_static
from .fit_tscan import fit_tscan
from .broadening import broadening
from .fit_irf import fit_irf
from .fit_seq import fit_seq
from .fit_eq import fit_eq
from .fit_osc import fit_osc

__all__ = ['broadening','fit_static', 'fit_tscan',
           'fit_irf', 'fit_seq', 'fit_eq', 'fit_osc']



'''
TRXASprefitpack: 
package for TRXAS pre fitting process which aims for the first order dynamics
TRXAS stands for Time Resolved X-ray Absorption Spectroscopy

:copyright: 2021-2022 by pistack (Junho Lee)
:license: LGPL3.
'''

from .mathfun import *

__version__ = '0.5.1'




'''
rate_eq:
submodule which solves 1st order rate equation and computes
the solution and signal

:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3.
'''

from typing import Optional, Tuple, Union
import numpy as np
import scipy.linalg as LA  # replace numpy.linalg to scipy.linalg
from .irf import calc_eta
from .A_matrix import make_A_matrix, make_A_matrix_cauchy
from .A_matrix import make_A_matrix_gau, make_A_matrix_pvoigt


def solve_model(equation: np.ndarray,
                y0: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:

    '''
    Solve system of first order rate equation

    Args:
      equation: matrix corresponding to model
      y0: initial condition

    Returns:
       1. eigenvalues of equation
       2. eigenvectors for equation
       3. coefficient where y0 = Vc
    '''

    eigval, V = LA.eig(equation)
    c = LA.solve(V, y0)

    return eigval.real, V, c
  
def solve_l_model(equation: np.ndarray,
                y0: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:

    '''
    Solve system of first order rate equation where the rate equation matrix is
    lower triangle

    Args:
      equation: matrix corresponding to model
      y0: initial condition

    Returns:
       1. eigenvalues of equation
       2. eigenvectors for equation
       3. coefficient where y0 = Vc
    '''

    eigval = np.diagonal(equation)
    V = np.eye(eigval.size)
    c = np.zeros(eigval.size)
    tmp = np.zeros(eigval.size)

    for i in range(1, eigval.size):
      tmp[:i] = eigval[:i]-eigval[i]
      tmp[:i][tmp[:i] == 0] = 1
      V[i, :i] = equation[i,:i] @ V[:i,:i]/tmp[:i]

    c[0] = y0[0]
    for i in range(1, eigval.size):
      c[i] = y0[i] - np.dot(c[:i], V[i,:i])

    return eigval, V, c

def solve_seq_model(tau):
    '''
    Solve sequential decay model
    
    sequential decay model: 
      0 -> 1 -> 2 -> 3 -> ... -> n 
    initial condition:
     y0 = [1, 0, 0, ..., 0] 

    Args:
      tau: liftime constants for each decay
      y0: initial condition

    Returns:
       1. eigenvalues of equation
       2. eigenvectors for equation
       3. coefficient to match initial condition
    '''
    eigval = np.zeros(tau.size+1)
    c = np.zeros(eigval.size)
    V = np.eye(eigval.size)
    
    eigval[:-1] = -1/tau

    for i in range(1, eigval.size):
      V[i, :i] = V[i-1,:i]*eigval[i-1]/(eigval[i]-eigval[:i])
    
    c[0] = 1
    for i in range(1, eigval.size):
      c[i] = -np.dot(c[:i], V[i,:i])
    return eigval, V, c

def compute_model(t: np.ndarray,
                  eigval: np.ndarray,
                  V: np.ndarray,
                  c: np.ndarray) -> np.ndarray:

    '''
    Compute solution of the system of rate equations solved by solve_model
    Note: eigval, V, c should be obtained from solve_model

    Args:
     t: time
     eigval: eigenvalue for equation
     V: eigenvectors for equation
     c: coefficient

    Returns:
      solution of rate equation

    Note:
      eigval, V, c should be obtained from solve_model.
    '''

    A = make_A_matrix(t, -eigval)
    y = (c * V) @ A
    return y


def compute_signal_gau(t: np.ndarray,
                       fwhm: float,
                       eigval: np.ndarray,
                       V: np.ndarray,
                       c: np.ndarray) -> np.ndarray:

    '''
    Compute solution of the system of rate equations solved by solve_model
    convolved with normalized gaussian distribution

    Args:
     t: time
     fwhm: full width at half maximum of normalized gaussian distribution
     eigval: eigenvalue for equation
     V: eigenvectors for equation
     c: coefficient

    Returns:
      Convolution of solution of rate equation and normalized gaussian
      distribution

    Note:
      eigval, V, c should be obtained from solve_model.
    '''

    A = make_A_matrix_gau(t, fwhm, -eigval)
    y_signal = (c * V) @ A
    return y_signal


def compute_signal_cauchy(t: np.ndarray,
                          fwhm: float,
                          eigval: np.ndarray,
                          V: np.ndarray,
                          c: np.ndarray) -> np.ndarray:

    '''
    Compute solution of the system of rate equations solved by solve_model
    convolved with normalized cauchy distribution

    Args:
     t: time
     fwhm: full width at half maximum of normalized cauchy distribution
     eigval: eigenvalue for equation
     V: eigenvectors for equation
     c: coefficient

    Returns:
      Convolution of solution of rate equation and normalized cauchy
      distribution

    Note:
      eigval, V, c should be obtained from solve_model.
    '''

    A = make_A_matrix_cauchy(t, fwhm, -eigval)
    y_signal = (c * V) @ A
    return y_signal


def compute_signal_pvoigt(t: np.ndarray,
                          fwhm_G: float,
                          fwhm_L: float,
                          eta: float,
                          eigval: np.ndarray,
                          V: np.ndarray,
                          c: np.ndarray) -> np.ndarray:

    '''
    Compute solution of the system of rate equations solved by solve_model
    convolved with normalized pseudo voigt profile

    .. math::

       \\mathrm{pvoigt}(t) = (1-\\eta) G(t) + \\eta L(t),

    G(t) stands for normalized gaussian,
    L(t) stands for normalized cauchy(lorenzian) distribution

    Args:
     t: time
     fwhm_G: full width at half maximum of gaussian part
     fwhm_L: full width at half maximum of cauchy part
     eta: mixing parameter
     eigval: eigenvalue for equation
     V: eigenvectors for equation
     c: coefficient

    Returns:
      Convolution of solution of rate equation and normalized pseudo
      voigt profile.

    Note:
      eigval, V, c should be obtained from solve_model.
    '''

    A = make_A_matrix_pvoigt(t, fwhm_G, fwhm_L, eta, -eigval)
    y_signal = (c * V) @ A
    return y_signal

def compute_signal_irf(t: np.ndarray, eigval: np.ndarray, V: np.ndarray, c: np.ndarray, 
fwhm: Union[float, np.ndarray], irf: Optional[str] = 'g', eta: Optional[float] = None):

  if irf == 'g':
    A = make_A_matrix_gau(t, fwhm, -eigval)

  elif irf == 'c':
    A = make_A_matrix_cauchy(t, fwhm, -eigval)

  elif irf == 'pv':

    if eta is None:
      eta = calc_eta(fwhm[0], fwhm[1])

    A = make_A_matrix_pvoigt(t, fwhm[0], fwhm[1], eta, -eigval)

  return (c * V) @ A

def fact_anal_model(model: np.ndarray, exclude: Optional[str] = None, 
data: Optional[np.ndarray] = None, eps: Optional[np.ndarray] = None):

  abs = np.zeros(model.shape[0])
  B = np.copy(model)
  y = np.copy(data)

  if eps is None:
    eps = np.ones_like(data)
  
  y = y/eps

  if exclude == 'first':
    C = np.einsum('j,ij->ij', 1/eps, B[1:, :])
  elif exclude == 'last':
    C = np.einsum('j,ij->ij', 1/eps, B[:-1,:])
  elif exclude == 'first_and_last':
    C = np.einsum('j,ij->ij', 1/eps, B[1:-1,:])
  else:
    C = np.einsum('j,ij->ij', 1/eps, B)
  
  coeff, _, _, _  = LA.lstsq(C.T, y)

  if exclude == 'first':
    abs[1:] = coeff
  elif exclude == 'last':
    abs[:-1] = coeff
  elif exclude == 'first_and_last':
    abs[1:-1] = coeff
  else:
    abs = coeff

  return abs

